@@ -12,8 +12,8 @@
 #include "midebugsession.h"
 
 #include "debuglog.h"
+#include "itoolviewfactoryholder.h"
 #include "midebugger.h"
-#include "midebuggerplugin.h"
 #include "mivariable.h"
 #include "mi/mi.h"
 #include "mi/micommand.h"
@@ -23,6 +23,7 @@
 #include <debugger/interfaces/iframestackmodel.h>
 #include <execute/iexecuteplugin.h>
 #include <interfaces/icore.h>
+#include <interfaces/idebugcontroller.h>
 #include <interfaces/idocument.h>
 #include <interfaces/idocumentcontroller.h>
 #include <interfaces/ilaunchconfiguration.h>
@@ -48,14 +49,47 @@ using namespace KDevMI::MI;
 
 namespace {
 constexpr DBGStateFlags notStartedDebuggerFlags{s_dbgNotStarted | s_appNotStarted};
+
+/**
+ * @return the string literal value of a given field of a given tuple value,
+ *         or an empty string if the tuple value does not contain the field
+ */
+[[nodiscard]] QString literalIfHasField(const MI::Value& tupleValue, const QString& fieldName)
+{
+    return tupleValue.hasField(fieldName) ? tupleValue[fieldName].literal() : QString{};
 }
 
-MIDebugSession::MIDebugSession(MIDebuggerPlugin *plugin)
+struct SignalInfo
+{
+    /**
+     * The value of a field "signal-name" of an MI record.
+     */
+    QString name;
+    /**
+     * The value of a field "signal-meaning" of an MI record or an empty string if the field is missing.
+     *
+     * @note GDB/MI always sends this field, e.g. "Aborted" for SIGABRT. LLDB-MI sends this field for
+     *       some signals (e.g. SIGSEGV) but does not send it for others (e.g. SIGABRT and SIGFPE).
+     */
+    QString meaning;
+
+    /**
+     * @return the values of fields "signal-name" and (optional) "signal-meaning" of a given MI record
+     */
+    [[nodiscard]] static SignalInfo fromRecord(const TupleRecord& record)
+    {
+        return {record[QStringLiteral("signal-name")].literal(),
+                literalIfHasField(record, QStringLiteral("signal-meaning"))};
+    }
+};
+
+} // unnamed namespace
+
+MIDebugSession::MIDebugSession()
     : m_procLineMaker(new ProcessLineMaker(this))
     , m_commandQueue(new CommandQueue)
     , m_debuggerState{notStartedDebuggerFlags}
     , m_tty(nullptr)
-    , m_plugin(plugin)
 {
     qCDebug(DEBUGGERCOMMON) << "Creating" << this;
 
@@ -65,6 +99,9 @@ MIDebugSession::MIDebugSession(MIDebuggerPlugin *plugin)
     connect(m_procLineMaker, &ProcessLineMaker::receivedStderrLines,
             this, &MIDebugSession::inferiorStderrLines);
 
+    connect(ICore::self()->debugController(), &IDebugController::currentSessionChanged, this,
+            &MIDebugSession::currentSessionChanged);
+
     // FIXME: see if this still works
     //connect(statusBarIndicator, SIGNAL(doubleClicked()),
     //        controller, SLOT(explainDebuggerStatus()));
@@ -274,7 +311,6 @@ bool MIDebugSession::startDebugging(const InferiorStartupInfo& startupInfo)
         m_commandQueue->clear();
         setDebuggerStateOn(s_shuttingDown);
         addGdbExitCommand();
-        emit reset();
         return false;
     }
 #endif
@@ -344,16 +380,32 @@ bool MIDebugSession::attachToProcess(int pid)
 
 void MIDebugSession::handleTargetAttach(const MI::ResultRecord& r)
 {
-    if (r.reason == QLatin1String("error")) {
-        const QString messageText =
-            i18n("<b>Could not attach debugger:</b><br />")+
-                 r[QStringLiteral("msg")].literal();
-        auto* message = new Sublime::Message(messageText, Sublime::Message::Error);
-        ICore::self()->uiController()->postMessage(message);
-        stopDebugger();
+    if (r.isReasonError()) {
+        stopDebuggerOnError(i18n("<b>Could not attach debugger:</b><br />") + r.errorMessage());
     }
 }
 
+void MIDebugSession::addFileExecAndSymbolsCommand(const QString& debuggeeExecutable)
+{
+    addCommand(MI::FileExecAndSymbols, debuggeeExecutable, this, &MIDebugSession::handleFileExecAndSymbols,
+               CmdHandlesError);
+}
+
+void MIDebugSession::handleFileExecAndSymbols(const ResultRecord& r)
+{
+    if (r.isReasonError()) {
+        stopDebuggerOnError(i18n("<b>Could not start debugger:</b><br />") + r.errorMessage());
+    }
+}
+
+void MIDebugSession::stopDebuggerOnError(const QString& errorMessage)
+{
+    Q_ASSERT(!errorMessage.isEmpty());
+    auto* const message = new Sublime::Message(errorMessage, Sublime::Message::Error);
+    ICore::self()->uiController()->postMessage(message);
+    stopDebugger();
+}
+
 bool MIDebugSession::examineCoreFile(const QUrl &debugee, const QUrl &coreFile)
 {
     emit showMessage(i18n("Examining core file %1", coreFile.toLocalFile()), 1000);
@@ -366,15 +418,36 @@ bool MIDebugSession::examineCoreFile(const QUrl &debugee, const QUrl &coreFile)
     }
 
     // FIXME: support non-local URLs
-    if (!loadCoreFile(nullptr, debugee.toLocalFile(), coreFile.toLocalFile())) {
-        return false;
+    const auto debugeePath = debugee.toLocalFile();
+
+    // If the file argument is not passed to the MI command -file-exec-and-symbols, GDB/MI clears the executable and
+    // symbol information. But since such information is absent at this time, sending the command is not useful.
+    // LLDB-MI does not support the absence of the file argument and replies to the bare
+    // -file-exec-and-symbols command with an error that terminates the debug session:
+    // 5^error,msg="Command 'file-exec-and-symbols'. Command Args. Validation failed. Mandatory args not found: file"
+    // Examining a core file works both with GDB and with LLDB if the command -file-exec-and-symbols is not sent at all.
+    if (!debugeePath.isEmpty()) {
+        addFileExecAndSymbolsCommand(debugeePath);
+        raiseEvent(connected_to_program);
     }
 
+    loadCoreFile(coreFile.toLocalFile());
     raiseEvent(program_state_changed);
 
     return true;
 }
 
+void MIDebugSession::initializeToolViewFactoryHolder(ToolViewFactoryHolderPtr&& holder)
+{
+    Q_ASSERT(!m_toolViewFactoryHolder);
+    m_toolViewFactoryHolder = std::move(holder);
+}
+
+auto MIDebugSession::takeToolViewFactoryHolder() -> ToolViewFactoryHolderPtr
+{
+    return std::move(m_toolViewFactoryHolder);
+}
+
 #define ENUM_NAME(o,e,v) (o::staticMetaObject.enumerator(o::staticMetaObject.indexOfEnumerator(#e)).valueToKey((v)))
 void MIDebugSession::setSessionState(DebuggerState state)
 {
@@ -384,9 +457,19 @@ void MIDebugSession::setSessionState(DebuggerState state)
     if (state != m_sessionState) {
         m_sessionState = state;
         emit stateChanged(state);
+        if (state == EndedState) {
+            // This session is no longer current. DebugController either has already switched
+            // away from the Debug sublime area or is not going to switch any time soon.
+            stopBeingCurrentSession();
+        }
     }
 }
 
+bool MIDebugSession::acceptsCommands() const
+{
+    return !debuggerStateIsOn(s_dbgNotStarted | s_shuttingDown);
+}
+
 bool MIDebugSession::debuggerStateIsOn(DBGStateFlags state) const
 {
     return m_debuggerState & state;
@@ -543,8 +626,6 @@ void MIDebugSession::stopDebugger()
             killDebuggerImpl();
         }
     });
-
-    emit reset();
 }
 
 void MIDebugSession::addGdbExitCommand()
@@ -697,11 +778,14 @@ void MIDebugSession::jumpToMemoryAddress(const QString& address)
 
 void MIDebugSession::addUserCommand(const QString& cmd)
 {
-    auto usercmd = createUserCommand(cmd);
-    if (!usercmd)
-        return;
+    auto adjustedCommand = cmd;
+    if (!cmd.isEmpty() && cmd[0].isDigit()) {
+        // Add a space to the beginning so debugger won't get confused if the
+        // command starts with a number (won't mix it up with command token added).
+        adjustedCommand.prepend(QLatin1Char{' '});
+    }
+    queueCmd(std::make_unique<UserCommand>(MI::NonMI, adjustedCommand));
 
-    queueCmd(std::move(usercmd));
     // User command can theoretically modify absolutely everything,
     // so need to force a reload.
 
@@ -712,17 +796,6 @@ void MIDebugSession::addUserCommand(const QString& cmd)
         raiseEvent(program_state_changed);
 }
 
-std::unique_ptr<MICommand> MIDebugSession::createUserCommand(const QString& cmd) const
-{
-    if (!cmd.isEmpty() && cmd[0].isDigit()) {
-        // Add a space to the beginning, so debugger won't get confused if the
-        // command starts with a number (won't mix it up with command token added)
-        return std::make_unique<UserCommand>(MI::NonMI, QLatin1Char(' ') + cmd);
-    } else {
-        return std::make_unique<UserCommand>(MI::NonMI, cmd);
-    }
-}
-
 std::unique_ptr<MICommand> MIDebugSession::createCommand(CommandType type, const QString& arguments,
                                                          CommandFlags flags) const
 {
@@ -929,6 +1002,39 @@ void MIDebugSession::raiseEvent(event_t e)
     }
 }
 
+void MIDebugSession::setCurrentPositionToFrameFieldOf(const TupleRecord& record)
+{
+    const auto& frame = record[QStringLiteral("frame")];
+    const auto file = literalIfHasField(frame, QStringLiteral("fullname"));
+    const auto line = literalIfHasField(frame, QStringLiteral("line"));
+    const auto addr = literalIfHasField(frame, QStringLiteral("addr"));
+
+    // An MI line is one-based and IDebugSession::currentLine() is zero-based. If line is an empty string,
+    // the value -1 that means "invalid, missing or unknown line number" is passed to setCurrentPosition().
+    setCurrentPosition(QUrl::fromLocalFile(file), line.toInt() - 1, addr);
+}
+
+void MIDebugSession::handleStackInfoFrame(const ResultRecord& result)
+{
+    setCurrentPositionToFrameFieldOf(result);
+}
+
+void MIDebugSession::askDebuggerAboutSelectedFrame()
+{
+    addCommand(StackInfoFrame, {}, this, &MIDebugSession::handleStackInfoFrame);
+}
+
+void MIDebugSession::coreFileLoaded()
+{
+    askDebuggerAboutSelectedFrame();
+    setDebuggerStateOn(s_programExited | s_core);
+}
+
+void MIDebugSession::failedToLoadCoreFile(const QString& debuggerErrorMessage)
+{
+    stopDebuggerOnError(i18n("<b>Failed to load core file:</b><br />") + debuggerErrorMessage);
+}
+
 bool KDevMI::MIDebugSession::hasCrashed() const
 {
     return m_hasCrashed;
@@ -1024,7 +1130,12 @@ void MIDebugSession::slotInferiorStopped(const MI::AsyncRecord& r)
     }
 
     if (reason == QLatin1String("exited-signalled")) {
-        programNoApp(i18n("Exited on signal %1", r[QStringLiteral("signal-name")].literal()));
+        const auto info = SignalInfo::fromRecord(r);
+        programNoApp(info.meaning.isEmpty()
+                         ? i18nc("%1 - the name of a POSIX signal", "Exited on signal %1", info.name)
+                         : i18nc("%1 - the name of a POSIX signal, %2 - a user-friendly name of the same signal",
+                                 "Exited on signal %1 (%2)", info.name, info.meaning));
+
         m_stateReloadNeeded = false;
         return;
     }
@@ -1044,14 +1155,13 @@ void MIDebugSession::slotInferiorStopped(const MI::AsyncRecord& r)
     bool wasInterrupt = false;
 
     if (reason == QLatin1String("signal-received")) {
-        QString name = r[QStringLiteral("signal-name")].literal();
-        QString user_name = r[QStringLiteral("signal-meaning")].literal();
+        const auto info = SignalInfo::fromRecord(r);
 
         // SIGINT is a "break into running program".
         // We do this when the user set/mod/clears a breakpoint but the
         // application is running.
         // And the user does this to stop the program also.
-        if (name == QLatin1String("SIGINT") && debuggerStateIsOn(s_interruptSent)) {
+        if (info.name == QLatin1String("SIGINT") && debuggerStateIsOn(s_interruptSent)) {
             wasInterrupt = true;
         } else {
             // Whenever we have a signal raised then tell the user, but don't
@@ -1059,40 +1169,30 @@ void MIDebugSession::slotInferiorStopped(const MI::AsyncRecord& r)
             // program has a signal that's caused the prog to stop.
             // Continuing from SIG FPE/SEGV will cause a "Cannot ..." and
             // that'll end the program.
-            programFinished(i18n("Program received signal %1 (%2)", name, user_name));
+            programFinished(info.meaning.isEmpty()
+                                ? i18nc("%1 - the name of a POSIX signal", "Program received signal %1", info.name)
+                                : i18nc("%1 - the name of a POSIX signal, %2 - a user-friendly name of the same signal",
+                                        "Program received signal %1 (%2)", info.name, info.meaning));
 
             m_hasCrashed = true;
         }
     }
 
-    if (!reason.contains(QLatin1String("exited"))) {
-        // FIXME: we should immediately update the current thread and
-        // frame in the framestackmodel, so that any user actions
-        // are in that thread. However, the way current framestack model
-        // is implemented, we can't change thread id until we refresh
-        // the entire list of threads -- otherwise we might set a thread
-        // id that is not already in the list, and it will be upset.
-
-        //Indicates if program state should be reloaded immediately.
-        bool updateState = false;
-
-        if (r.hasField(QStringLiteral("frame"))) {
-            const MI::Value& frame = r[QStringLiteral("frame")];
-            QString file, line, addr;
-
-            if (frame.hasField(QStringLiteral("fullname"))) file = frame[QStringLiteral("fullname")].literal();
-            if (frame.hasField(QStringLiteral("line")))     line = frame[QStringLiteral("line")].literal();
-            if (frame.hasField(QStringLiteral("addr")))     addr = frame[QStringLiteral("addr")].literal();
-
-            // gdb counts lines from 1 and we don't
-            setCurrentPosition(QUrl::fromLocalFile(file), line.toInt() - 1, addr);
-
-            updateState = true;
-        }
-
-        if (updateState) {
-            reloadProgramState();
-        }
+    // FIXME: we should immediately update the current thread and
+    // frame in the framestackmodel, so that any user actions
+    // are in that thread. However, the way current framestack model
+    // is implemented, we can't change thread id until we refresh
+    // the entire list of threads -- otherwise we might set a thread
+    // id that is not already in the list, and it will be upset.
+    if (r.hasField(QStringLiteral("frame"))) {
+        setCurrentPositionToFrameFieldOf(r);
+        reloadProgramState();
+    } else {
+        qCDebug(DEBUGGERCOMMON) << "a non-exited *stopped MI output record does not have a field \"frame\"";
+        // Ask about selected frame after reloading the program state so that correct values
+        // of the --thread and --frame arguments are added to the command -stack-info-frame.
+        reloadProgramState();
+        askDebuggerAboutSelectedFrame();
     }
 
     setDebuggerStateOff(s_interruptSent);
@@ -1213,19 +1313,15 @@ void MIDebugSession::explainDebuggerStatus()
 // FIXME: connect to debugger's slot.
 void MIDebugSession::defaultErrorHandler(const MI::ResultRecord& result)
 {
-    QString msg = result[QStringLiteral("msg")].literal();
+    const auto errorMessage = result.errorMessage();
 
-    if (msg.contains(QLatin1String("No such process")))
-    {
+    if (errorMessage.contains(QLatin1String("No such process"))) {
         setDebuggerState(s_appNotStarted|s_programExited);
         raiseEvent(program_exited);
         return;
     }
 
-    const QString messageText =
-        i18n("<b>Debugger error</b>"
-             "<p>Debugger reported the following error:"
-             "<p><tt>%1", result[QStringLiteral("msg")].literal());
+    const QString messageText = i18n("<b>Debugger error:</b><br />") + errorMessage;
     auto* message = new Sublime::Message(messageText, Sublime::Message::Error);
     ICore::self()->uiController()->postMessage(message);
 
@@ -1250,4 +1346,27 @@ void MIDebugSession::setSourceInitFile(bool enable)
     m_sourceInitFile = enable;
 }
 
+bool MIDebugSession::isCurrentDebugSession(const QObject* object)
+{
+    return object == ICore::self()->debugController()->currentSession();
+}
+
+void MIDebugSession::currentSessionChanged(IDebugSession* session, IDebugSession* previousSession)
+{
+    if (session && previousSession == this) {
+        // another debug session replaced this one as current
+        stopBeingCurrentSession();
+    }
+    // else one of the following statements is true:
+    // * some other session stopped being current (probably *this* session just became current);
+    // * this session just ended and DebugController is about to switch away from
+    //   the Debug sublime area => setSessionState() will stop being current a bit later.
+}
+
+void MIDebugSession::stopBeingCurrentSession()
+{
+    // maintain the invariant that a non-current session does not hold tool view factories
+    m_toolViewFactoryHolder.reset();
+}
+
 #include "moc_midebugsession.cpp"
