@@ -9,6 +9,8 @@
 
 #include "debuggertestbase.h"
 
+#include "mi/mi.h"
+#include "mi/micommand.h"
 #include "midebugsession.h"
 #include "testhelper.h"
 
@@ -36,6 +38,7 @@
 #include <QDebug>
 #include <QFileInfo>
 #include <QMetaObject>
+#include <QPointer>
 #include <QSignalBlocker>
 #include <QSignalSpy>
 #include <QStandardPaths>
@@ -46,6 +49,7 @@
 
 #include <algorithm>
 #include <array>
+#include <memory>
 
 using namespace KDevelop;
 using namespace KDevMI;
@@ -142,6 +146,15 @@ void verifyCurrentLocation(const IDebugSession* session, const QUrl& url, int mi
             QCOMPARE_EQ(stackModel->rowCount(threadIndex), expectedFrameCount);                                        \
     } while (false)
 
+class MICommandResultSpy : public QObject, public QStringList
+{
+public:
+    void handle(const MI::ResultRecord& record)
+    {
+        push_back(record["value"].literal());
+    }
+};
+
 } // unnamed namespace
 
 ICore* DebuggerTestBase::core() const
@@ -1395,12 +1408,55 @@ void DebuggerTestBase::testStackSwitchThread()
     WAIT_FOR_STATE(session, IDebugSession::EndedState);
 }
 
-void DebuggerTestBase::testCoreFile()
+void DebuggerTestBase::testThreadAndFrameInfo()
+{
+    // Verify that `--thread <current thread number>` is added to commands that can specify a thread but do not do so.
+
+    auto* const session = createTestDebugSession();
+    TestLaunchConfiguration cfg("debuggee_debugeethreads");
+    const auto fileName = findSourceFile("debugeethreads.cpp");
+
+    breakpoints()->addCodeBreakpoint(QUrl::fromLocalFile(fileName), 43); // QThread::msleep(600);
+    const ActiveStateSessionSpy sessionSpy(session);
+    START_DEBUGGING_AND_WAIT_FOR_PAUSED_STATE_E(session, cfg, sessionSpy);
+
+    const QSignalSpy userCommandOutputSpy(session, &MIDebugSession::debuggerUserCommandOutput);
+
+    session->addCommand(std::make_unique<MI::UserCommand>(MI::ThreadInfo, ""));
+    session->addCommand(std::make_unique<MI::UserCommand>(MI::StackListLocals, "0"));
+    WAIT_FOR_STATE_AND_IDLE(session, IDebugSession::PausedState); // wait for command finish
+
+    // The user command outputs should be:
+    // 0. -thread-info
+    // 1. ^done for thread-info
+    // 2. -stack-list-locals
+    // 3. ^done for -stack-list-locals
+    QCOMPARE(userCommandOutputSpy.count(), 4);
+    QVERIFY(userCommandOutputSpy.at(2).constFirst().toString().contains("--thread 1"));
+
+    session->run();
+    WAIT_FOR_STATE(session, IDebugSession::EndedState);
+}
+
+/**
+ * @return the name of a core dump file
+ *
+ * @param reuseExisting whether to reuse an existing file at the location of a core file;
+ *        if @c false, the existing file (if any) is removed and a new core dump is generated
+ *
+ * Call RETURN_IF_TEST_RESOLVED() after this function.
+ */
+[[nodiscard]] static QString generateCoreFile(bool reuseExisting)
 {
     QFileInfo f("core");
     f.setCaching(false); // don't cache information
     if (f.exists()) {
-        QVERIFY(QFile::remove(f.canonicalFilePath()));
+        if (reuseExisting) {
+            qDebug() << "reusing an existing core file";
+            return f.canonicalFilePath();
+        }
+        qDebug() << "removing an existing core file to generate a new one";
+        QVERIFY_RETURN(QFile::remove(f.canonicalFilePath()), QString{});
     }
 
     KProcess debugeeProcess;
@@ -1424,24 +1480,185 @@ void DebuggerTestBase::testCoreFile()
         }
     }
     if (!coreFileFound) {
-        QSKIP("no core dump found, check your system configuration (see /proc/sys/kernel/core_pattern).");
+        QSKIP_RETURN("No core dump found, check your system configuration (see /proc/sys/kernel/core_pattern)",
+                     QString{});
     }
 
+    return f.canonicalFilePath();
+}
+
+void DebuggerTestBase::testCoreFile_data()
+{
+    QTest::addColumn<FileKind>("executableFileKind");
+    QTest::addColumn<FileKind>("coreFileKind");
+
+    const auto addRow = [](FileKind executableFileKind, FileKind coreFileKind) {
+        QTest::addRow("executable-%s,core-%s", enumeratorName(executableFileKind), enumeratorName(coreFileKind))
+            << executableFileKind << coreFileKind;
+    };
+
+    // Save time: do not test the combinations where neither executable nor core file kind is acceptable.
+    for (const auto executableFileKind : allFileKinds) {
+        if (isAcceptableExecutableFileKindForCore(executableFileKind)) {
+            for (const auto coreFileKind : allFileKinds) {
+                addRow(executableFileKind, coreFileKind);
+            }
+        } else {
+            addRow(executableFileKind, acceptableCoreFileKind);
+        }
+    }
+
+    // Add a single representative smoke test for two unacceptable file kinds.
+    addRow(FileKind::Invalid, FileKind::Nonexistent);
+}
+
+void DebuggerTestBase::testCoreFile()
+{
+    QFETCH(const FileKind, executableFileKind);
+    QFETCH(const FileKind, coreFileKind);
+
+    const auto findDebuggeeExecutable = [] {
+        return findExecutable("debuggee_crash");
+    };
+    const auto generateCoreFile = [this] {
+        // If a core file was just generated for another data row, it is up-to-date and can be reused.
+        // Otherwise, generate a new core file in place of a possibly incompatible existing one.
+        const auto reuseExisting = m_generatedCoreFile;
+        const auto fileName = ::generateCoreFile(reuseExisting);
+        m_generatedCoreFile = !fileName.isEmpty();
+        return QUrl::fromLocalFile(fileName);
+    };
+
+    const auto executable = urlForFileKind(executableFileKind, findDebuggeeExecutable);
+    RETURN_IF_TEST_RESOLVED();
+    const auto core = urlForFileKind(coreFileKind, generateCoreFile);
+    RETURN_IF_TEST_RESOLVED();
+
     auto* const session = createTestDebugSession();
-    session->examineCoreFile(findExecutable("debuggee_crash"), QUrl::fromLocalFile(f.canonicalFilePath()));
+    VERIFY_INVALID_CURRENT_LOCATION(session);
+    QVERIFY(session->examineCoreFile(executable, core));
 
-    const auto* const stackModel = session->frameStackModel();
+    const auto isExecutableAcceptable = isAcceptableExecutableFileKindForCore(executableFileKind);
+    if (isExecutableAcceptable && coreFileKind == acceptableCoreFileKind) {
+        const auto* const stackModel = session->frameStackModel();
 
-    WAIT_FOR_STATE(session, IDebugSession::StoppedState);
+        WAIT_FOR_STATE(session, IDebugSession::StoppedState);
 
-    const auto threadIndex = stackModel->index(0, 0);
-    VALIDATE_COLUMN_COUNTS_THREAD_COUNT_AND_STACK_FRAME_NUMBERS(threadIndex, 1);
-    COMPARE_DATA(threadIndex, adjustedStackModelFrameName("#1 at foo"));
+        const auto missingCurrentLocation =
+            executableFileKind == FileKind::EmptyName && !isLldb() && session->currentUrl().isEmpty();
+        if (missingCurrentLocation) {
+            // No idea why one day GDB loads the current location from the core file, then 10 days later it does not...
+            qWarning() << "GDB failed to load current location from the core file";
+        } else {
+            VERIFY_VALID_CURRENT_LOCATION(session, QUrl::fromLocalFile(findSourceFile("debugeecrash.cpp")), 25);
+        }
+
+        const auto threadIndex = stackModel->index(0, 0);
+        VALIDATE_COLUMN_COUNTS_THREAD_COUNT_AND_STACK_FRAME_NUMBERS(threadIndex, 1);
+        if (!missingCurrentLocation) {
+            COMPARE_DATA(threadIndex, adjustedStackModelFrameName("#1 at foo"));
+        }
+
+        session->stopDebugger();
+    } else if (isExecutableAcceptable && coreFileKind == FileKind::EmptyName && !isLldb()) {
+        // If the core filename is empty, KDevelop sends a non-MI command `core` without argument, GDB
+        // outputs a message "No core file now." and replies with "done". Consequently, the debug session
+        // is stuck in the starting state instead of ending with an error. This is not a problem in
+        // practice, because the Select Core File dialog cannot be accepted if the core filename is empty.
+        WAIT_FOR_A_WHILE(session, 1000);
+        QEXPECT_FAIL("", "Debug session is stuck in the starting state", Continue);
+        QCOMPARE(session->state(), IDebugSession::EndedState);
+        QCOMPARE(session->state(), IDebugSession::StartingState);
+        session->stopDebugger();
+    } // else: the debug session ends with an error => nothing to do other than verify the ending
 
-    session->stopDebugger();
     WAIT_FOR_STATE(session, IDebugSession::EndedState);
 }
 
+void DebuggerTestBase::testCommandHandler_data()
+{
+    QTest::addColumn<bool>("addCommandWithCurrentSessionHandler");
+    QTest::addColumn<bool>("destroyHandler");
+    QTest::addColumn<bool>("replaceSession");
+
+    for (const auto addCommandWithCurrentSessionHandler : {false, true}) {
+        for (const auto destroyHandler : {false, true}) {
+            for (const auto replaceSession : {false, true}) {
+                QTest::addRow("%s-%s-%s",
+                              addCommandWithCurrentSessionHandler ? "addCommandWithCurrentSessionHandler"
+                                                                  : "addCommand",
+                              destroyHandler ? "destroy_handler" : "", replaceSession ? "replace_session" : "")
+                    << addCommandWithCurrentSessionHandler << destroyHandler << replaceSession;
+            }
+        }
+    }
+}
+
+void DebuggerTestBase::testCommandHandler()
+{
+    QFETCH(const bool, addCommandWithCurrentSessionHandler);
+    QFETCH(const bool, destroyHandler);
+    QFETCH(const bool, replaceSession);
+
+    const QPointer session{createTestDebugSession()};
+    TestLaunchConfiguration cfg;
+
+    constexpr auto breakpointLine = 40; // return 0;
+    addDebugeeBreakpoint(breakpointLine);
+
+    const ActiveStateSessionSpy sessionSpy(session);
+    START_DEBUGGING_AND_WAIT_FOR_PAUSED_STATE_E(session, cfg, sessionSpy);
+    QCOMPARE(currentMiLine(session), breakpointLine);
+
+    auto handler = std::make_unique<MICommandResultSpy>();
+
+    const auto addCommand = [addCommandWithCurrentSessionHandler](MIDebugSession& session, const auto&... args) {
+        if (addCommandWithCurrentSessionHandler) {
+            session.addCommandWithCurrentSessionHandler(args...);
+        } else {
+            session.addCommand(args...);
+        }
+    };
+    addCommand(*session, MI::DataEvaluateExpression, "ts.a", handler.get(), &MICommandResultSpy::handle);
+
+    // NOTE: the run() command (continue execution) is removed from the command queue and never sent to the debugger if
+    //       another session is created below (because starting a new session stops the debugger of the previous one).
+    session->run();
+    QCOMPARE(handler->size(), 0); // an MI command is asynchronous
+
+    if (destroyHandler) {
+        handler.reset();
+    }
+
+    if (replaceSession) {
+        auto* const session2 = createTestDebugSession();
+
+        const ActiveStateSessionSpy sessionSpy2(session2);
+        START_DEBUGGING_AND_WAIT_FOR_PAUSED_STATE_E(session2, cfg, sessionSpy2);
+        QCOMPARE(currentMiLine(session2), breakpointLine);
+
+        session2->run();
+        WAIT_FOR_STATE(session2, IDebugSession::EndedState);
+    }
+
+    // Wait for the (first) session to finish and be destroyed (if it is still alive).
+    QTRY_COMPARE(session, nullptr);
+
+    QCOMPARE(static_cast<bool>(handler), !destroyHandler);
+    if (destroyHandler) {
+        return; // no crash - good, nothing else to verify
+    }
+
+    if (addCommandWithCurrentSessionHandler && replaceSession) {
+        // The first session stopped being current before the result of the command
+        // arrived, therefore the current-session handler must not be invoked.
+        QCOMPARE(handler->size(), 0);
+    } else {
+        QCOMPARE(handler->size(), 1);
+        QCOMPARE(handler->constFirst(), "1");
+    }
+}
+
 void DebuggerTestBase::testVariablesLocalsStruct()
 {
     auto* const session = createTestDebugSession();
@@ -1808,24 +2025,123 @@ void DebuggerTestBase::testCommandOrderFastStepping()
     WAIT_FOR_STATE(session, IDebugSession::EndedState);
 }
 
-void DebuggerTestBase::testSegfaultDebugee()
+void DebuggerTestBase::testReceivePosixSignal_data()
 {
+    QTest::addColumn<QString>("executableName");
+    QTest::addColumn<QString>("sourceFileName");
+    QTest::addColumn<QString>("signalName");
+    QTest::addColumn<bool>("isSignaRaisedInLibraryFunction");
+
+    QTest::newRow("debuggee_crash") << "debuggee_crash" << "debugeecrash.cpp" << "SIGSEGV" << false;
+    QTest::newRow("debuggee_abort") << "debuggee_abort" << "debugeeabort.cpp" << "SIGABRT" << true;
+    QTest::newRow("debuggee_arithmetic_signal")
+        << "debuggee_arithmetic_signal" << "debugeearithmeticsignal.cpp" << "SIGFPE" << false;
+}
+
+void DebuggerTestBase::testReceivePosixSignal()
+{
+    QFETCH(const QString, executableName);
+    QFETCH(const QString, sourceFileName);
+
     auto* const session = createTestDebugSession();
-    TestLaunchConfiguration cfg("debuggee_crash");
-    const auto fileName = findSourceFile("debugeecrash.cpp");
+    TestLaunchConfiguration cfg(executableName);
+    const auto sourceFilePath = findSourceFile(sourceFileName);
+    const auto sourceUrl = QUrl::fromLocalFile(sourceFilePath);
 
     session->variableController()->setAutoUpdate(IVariableController::UpdateLocals);
 
-    breakpoints()->addCodeBreakpoint(QUrl::fromLocalFile(fileName), 23);
+    const auto* const stackModel = session->frameStackModel();
+
+    breakpoints()->addCodeBreakpoint(sourceUrl, 23);
     ActiveStateSessionSpy sessionSpy(session);
     START_DEBUGGING_AND_WAIT_FOR_PAUSED_STATE_E(session, cfg, sessionSpy);
     QCOMPARE(session->currentLine(), 23);
 
-    CONTINUE_AND_WAIT_FOR_PAUSED_STATE(session, sessionSpy);
-    QCOMPARE(currentMiLine(session), 25);
+    QSignalSpy errorOutputSpy(session, &MIDebugSession::inferiorStderrLines);
 
-    session->stopDebugger();
+    CONTINUE_AND_WAIT_FOR_PAUSED_STATE(session, sessionSpy);
+
+    QVERIFY(session->hasCrashed());
+
+    QFETCH(const bool, isSignaRaisedInLibraryFunction);
+    constexpr auto signalLine = 25;
+
+    if (isSignaRaisedInLibraryFunction) {
+        VERIFY_VALID_ADDRESS(session->currentAddr());
+        QCOMPARE_NE(session->currentUrl(), sourceUrl);
+    } else {
+        VERIFY_VALID_CURRENT_LOCATION(session, sourceUrl, signalLine);
+    }
+
+    const auto threadIndex = stackModel->index(0, 0);
+    auto firstDebuggeeFrameNumber = 0;
+    if (isSignaRaisedInLibraryFunction) {
+        firstDebuggeeFrameNumber = -1;
+        for (auto frame = 0, count = stackModel->rowCount(threadIndex); frame < count; ++frame) {
+            // The line signalLine is in the function main() of the debuggee.
+            if (stackModel->index(frame, 1, threadIndex).data().toString() == "main") {
+                firstDebuggeeFrameNumber = frame;
+                break;
+            }
+        }
+        QCOMPARE_GT(firstDebuggeeFrameNumber, 0); // the frame #0 is in a library function
+    }
+    COMPARE_DATA(stackModel->index(firstDebuggeeFrameNumber, 2, threadIndex),
+                 sourceFilePath + ':' + QString::number(signalLine));
+
+    const auto takeStderrOutput = [&errorOutputSpy] {
+        QStringList errorOutputLines;
+        for (const auto& arguments : std::as_const(errorOutputSpy)) {
+            QCOMPARE_RETURN(arguments.size(), 1, QString{}); // MIDebugSession::inferiorStderrLines() has one parameter
+            errorOutputLines << arguments.constFirst().toStringList();
+        }
+        errorOutputSpy.clear();
+
+        auto ret = errorOutputLines.join('\n');
+        if (errorOutputLines.empty()) {
+            qDebug() << "inferior stderr output is empty";
+        } else {
+            qDebug().noquote() << "inferior stderr output:\n" << ret;
+        }
+        return ret;
+    };
+    QFETCH(const QString, signalName);
+
+    auto stderrOutput = takeStderrOutput();
+    RETURN_IF_TEST_FAILED();
+    QVERIFY(stderrOutput.contains(signalName));
+
+    session->run();
     WAIT_FOR_STATE(session, IDebugSession::EndedState);
+
+    stderrOutput = takeStderrOutput();
+    RETURN_IF_TEST_FAILED();
+    // An MI output record like the following is expected:
+    // *stopped,reason="exited-signalled",signal-name="SIGSEGV",signal-meaning="Segmentation fault"
+    // Verify that the received signal name is present in a final message
+    // appended by KDevelop to the stderr output of the inferior.
+    if (isLldb()) {
+        QEXPECT_FAIL("",
+                     "The reason of the *stopped MI output record sent by LLDB-MI "
+                     "is \"exited-normally\" instead of \"exited-signalled\"",
+                     Continue);
+    } else if (stderrOutput.contains("01")) {
+        // In case of a SIGSEGV or a SIGFPE signal, GDB/MI sends *stopped,reason="exited",exit-code="01" in place of
+        // the expected *stopped MI output record with reason="exited-signalled" on the openSUSE Tumbleweed CI servers.
+        // KDevelop then appends to the stderr output of the inferior a different final message that contains the
+        // received exit code. GDB/MI sends the expected output record on developers' local GNU/Linux systems and on the
+        // FreeBSD CI server. Work around the difference by accepting either the expected signal name or the exit code.
+        // TODO: why does this difference in output records exists despite equal GDB version numbers?
+        //       Is this some openSUSE bug or a CI server configuration issue?
+        if (QTest::currentDataTag() == QLatin1String{"debuggee_crash"}
+            || QTest::currentDataTag() == QLatin1String{"debuggee_arithmetic_signal"}) {
+            QEXPECT_FAIL("",
+                         "The reason of the *stopped MI output record sent by GDB/MI "
+                         "is \"exited\" instead of \"exited-signalled\"",
+                         Continue);
+        }
+    }
+    QVERIFY(stderrOutput.contains(signalName));
 }
 
 void DebuggerTestBase::testDebugInExternalTerminal_data()
@@ -1883,4 +2199,39 @@ QString DebuggerTestBase::adjustedStackModelFrameName(QString frameName) const
     return frameName;
 }
 
+bool DebuggerTestBase::isAcceptableExecutableFileKindForCore(FileKind executableFileKind)
+{
+    return executableFileKind == FileKind::Valid || executableFileKind == FileKind::EmptyName;
+}
+
+template<typename UrlProvider>
+QUrl DebuggerTestBase::urlForFileKind(FileKind fileKind, UrlProvider validUrlCallback) const
+{
+    switch (fileKind) {
+    case FileKind::Valid:
+        return validUrlCallback();
+    case FileKind::EmptyName:
+        return QUrl{};
+    case FileKind::Nonexistent:
+        return QUrl::fromLocalFile("/this/path/should/not/exist");
+    case FileKind::Directory:
+        return debugeeUrl().adjusted(QUrl::RemoveFilename | QUrl::StripTrailingSlash);
+    case FileKind::NotReadable: {
+        if (m_notReadableFile.fileName().isEmpty()) {
+            QVERIFY_RETURN(m_notReadableFile.open(), QUrl{});
+            m_notReadableFile.close();
+            QCOMPARE_RETURN(m_notReadableFile.error(), QFileDevice::NoError, QUrl{});
+            QVERIFY_RETURN(m_notReadableFile.setPermissions({}), QUrl{});
+        }
+        const auto fileName = m_notReadableFile.fileName();
+        QVERIFY_RETURN(!fileName.isEmpty(), QUrl{});
+        QVERIFY_RETURN(!QFileInfo{fileName}.isReadable(), QUrl{});
+        return QUrl::fromLocalFile(fileName);
+    }
+    case FileKind::Invalid:
+        return debugeeUrl(); // a C++ source file is neither an executable nor a core dump file
+    }
+    Q_UNREACHABLE_RETURN(QUrl{});
+}
+
 #include "moc_debuggertestbase.cpp"
